/*
 *  linux/drivers/char/pcxe.c
 * 
 *  Written by Troy De Jongh, November, 1994
 *
 *  Copyright (C) 1994 Troy De Jongh
 *  This software may be used and distributed according to the terms 
 *  of the GNU Public License.
 *
 *  This driver is for the DigiBoard PC/Xe and PC/Xi line of products.
 *
 *  This driver does NOT support DigiBoard's fastcook FEP option and
 *  does not support the transparent print (i.e. digiprint) option.
 *
 *  Please email any suggestions or bug reports to troyd@skypoint.com
 *
 */

#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/major.h>
#include <linux/delay.h>
#include <linux/serial.h>
#include <linux/tty_driver.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/bitops.h>

#define VERSION "1.0b"     /* Beta */

#include "digi.h"
#include "fep.h"
#include "pcxx.h"
#include "digi_fep.h"
#include "digi_bios.h"
#include "pcxxconfig.h"

#define DIGIMAJOR   30
#define DIGICUMAJOR 31 

#define MAXPORTS 16     /* Max ports per PC/Xx type board */
#define NBDEVS   (NUMCARDS * MAXPORTS)
#define PORTNUM(x)  ((x)->dev % MAXPORTS)

int pcxx_ncook=sizeof(pcxx_cook);
int pcxx_nbios=sizeof(pcxx_bios);

struct channel digi_channels[NBDEVS];

#define MIN(a,b)	((a) < (b) ? (a) : (b))
#define pcxxassert(x, msg)  if(!(x)) pcxx_error(__LINE__, msg)

#define FEPTIMEOUT 200000  
#define SERIAL_TYPE_NORMAL	1
#define SERIAL_TYPE_CALLOUT	2
#define PCXE_EVENT_HANGUP   1

static unsigned char *pcxx_buf;
static struct semaphore pcxx_sem = MUTEX;

static char mesg[100];

struct tty_driver pcxe_driver;
struct tty_driver pcxe_callout;
static struct timer_list pcxe_timer = {NULL, NULL, 0, 0, 0};
static int pcxe_refcount;

DECLARE_TASK_QUEUE(tq_pcxx);
static struct tty_struct *pcxe_table[NBDEVS];
static struct termios *pcxe_termios[NBDEVS];
static struct termios *pcxe_termios_locked[NBDEVS];

static void pcxxpoll(unsigned long);
static void pcxxdelay(int);
static void fepcmd(struct channel *, int, int, int, int, int);
static void pcxe_put_char(struct tty_struct *, unsigned char);
static void pcxe_flush_chars(struct tty_struct *);
static void pcxx_error(int, char *);
static void pcxe_close(struct tty_struct *, struct file *);
static int pcxe_ioctl(struct tty_struct *, struct file *, unsigned int, unsigned long);
static void pcxe_set_termios(struct tty_struct *, struct termios *);
static int pcxe_write(struct tty_struct *, int, unsigned char *, int);
static int pcxe_write_room(struct tty_struct *);
static int pcxe_chars_in_buffer(struct tty_struct *);
static void pcxe_flush_buffer(struct tty_struct *);
static void doevent(int);
static void data_to_board(struct channel *, unchar *, int);
static void receive_data(struct channel *);
static void pcxxparam(struct tty_struct *);
static void do_softint(void *);
static inline void pcxe_sched_event(struct channel *, int);
static void do_pcxe_bh(void *);
static void pcxe_start(struct tty_struct *);
static void pcxe_stop(struct tty_struct *);
static void digi_send_break(struct channel *ch, int msec);
static void shutdown(struct channel *);
void console_print(const char *);
void setup_empty_event(struct tty_struct *tty);
static inline void memwinon(struct board_info *b, unsigned int win);
static inline void memwinoff(struct board_info *b, unsigned int win);
static inline void globalwinon(struct channel *ch);
static inline void rxwinon(struct channel *ch);
static inline void txwinon(struct channel *ch);
static inline void memoff(struct channel *ch);
static inline void assertgwinon(struct channel *ch);
static inline void assertmemoff(struct channel *ch);

/* These inline routines are to turn board memory on and off */
static inline void memwinon(struct board_info *b, unsigned int win)
{
	if(b->type == PCXEVE)
		outb_p(FEPWIN|win, b->port+1);
	else
		outb_p(inb(b->port)|FEPMEM, b->port);
}

static inline void memwinoff(struct board_info *b, unsigned int win)
{
	outb_p(inb(b->port)&~FEPMEM, b->port);
	if(b->type == PCXEVE)
		outb_p(0, b->port + 1);
}

static inline void globalwinon(struct channel *ch)
{
	if(ch->board->type == PCXEVE)
		outb_p(FEPWIN, ch->board->port+1);
	else
		outb_p(FEPMEM, ch->board->port);
}

static inline void rxwinon(struct channel *ch)
{
	if(ch->rxwin == 0)
		outb_p(FEPMEM, ch->board->port);
	else 
		outb_p(ch->rxwin, ch->board->port+1);
}

static inline void txwinon(struct channel *ch)
{
	if(ch->txwin == 0)
		outb_p(FEPMEM, ch->board->port);
	else
		outb_p(ch->txwin, ch->board->port+1);
}

static inline void memoff(struct channel *ch)
{
	outb_p(0, ch->board->port);
	if(ch->board->type == PCXEVE)
		outb_p(0, ch->board->port+1);
}

static inline void assertgwinon(struct channel *ch)
{
	if(ch->board->type != PCXEVE)
		pcxxassert(inb(ch->board->port) & FEPMEM, "Global memory off");
}

static inline void assertmemoff(struct channel *ch)
{
	if(ch->board->type != PCXEVE)
		pcxxassert(!(inb(ch->board->port) & FEPMEM), "Memory on");
}

static inline void pcxe_sched_event(struct channel *info, int event)
{
	info->event |= 1 << event;
	queue_task_irq_off(&info->tqueue, &tq_pcxx);
	mark_bh(DIGI_BH);
}

static void pcxx_error(int line, char *msg)
{
	sprintf(mesg, "pcxx_error (DigiBoard): line=%d %s\n", line, msg);
	console_print(mesg);
	return;
}


static void pcxe_close(struct tty_struct * tty, struct file * filp)
{
	struct channel *ch, *info;
	struct board_chan *bc;
	int line;

	/*
	 * I think it's rather weird that Linux's TTY subsystem
	 * calls close() even though the open() just failed.  That
	 * means I have to check for a failed open() here also.
	 */
	line = MINOR(tty->device) - tty->driver.minor_start;
	if(line < 0 || line >= NBDEVS) {
		console_print("line out of range in pcxe_close\n");
		tty->driver_data = NULL;
		return;
	}

	if(tty->driver_data == NULL)    /* open() must've failed earlier...*/
		return;

	if(tty_hung_up_p(filp))
		return;

	ch = info = (struct channel *) tty->driver_data;
	bc = ch->brdchan;

	if((tty->count == 1) && (info->count != 1)) {
		printk("pcxe_close: bad serial port count; tty->count is 1, "
		       "info->count is %d\n", info->count);
		info->count = 1;
	}
	if(--info->count < 0) {
		printk("pcxe_close: bad serial port count, minor=%d\n", 
					MINOR(tty->device));
		info->count = 0;
	}
	if(info->count)
		return;
	info->asyncflags |= ASYNC_CLOSING;

	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	if(info->asyncflags & ASYNC_NORMAL_ACTIVE)
		info->normal_termios = *tty->termios;
	if(info->asyncflags & ASYNC_CALLOUT_ACTIVE)
		info->callout_termios = *tty->termios;
	if(info->asyncflags & ASYNC_INITIALIZED) {
		setup_empty_event(tty);		
		wait_until_sent(tty, 3000); /* 30 seconds timeout */
	}

	shutdown(info);
	if(tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
	if(tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
	info->event = 0;


	if(tty->ldisc.num != ldiscs[N_TTY].num) {
		if(tty->ldisc.close)
			(tty->ldisc.close)(tty);
		tty->ldisc = ldiscs[N_TTY];
		tty->termios->c_line = N_TTY;
		if(tty->ldisc.open)
			(tty->ldisc.open)(tty);
	}
	if(info->blocked_open) {
		if(info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + info->close_delay;
			schedule();
		}
		wake_up_interruptible(&info->open_wait);
	}
	info->asyncflags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|
			 ASYNC_CLOSING);
	wake_up_interruptible(&info->close_wait);
}

static void shutdown(struct channel *info)
{
	unsigned long flags;
	struct board_chan *bc;
	struct tty_struct *tp;

	if(!(info->asyncflags & ASYNC_INITIALIZED))
		return;

	save_flags(flags);
	cli();
	globalwinon(info);

	bc = info->brdchan;

	if(bc)
		bc->idata = 0;

	tp = info->tty;

	/*
	 * If we're a modem control device and HUPCL is on, drop RTS & DTR.
	 */
	if((info->asyncflags & ASYNC_CALLOUT_ACTIVE) && tp->termios->c_cflag & HUPCL) {
		info->omodem &= ~(RTS|DTR);
		fepcmd(info, SETMODEM, 0, DTR|RTS, 10, 1);
	}

	memoff(info);
	restore_flags(flags);

	/* Hold RTS and DTR low for a 1/3 of a second so we have a sure hangup... */
	current->state = TASK_INTERRUPTIBLE;
	current->timeout = jiffies + HZ/3;
	schedule();

	info->asyncflags &= ~ASYNC_INITIALIZED;
	return;
}


void pcxe_hangup(struct tty_struct *tty)
{
	struct channel *ch;

	ch = (struct channel *) tty->driver_data;
	
	shutdown(ch);
	ch->event = 0;
	ch->count = 0;
	ch->asyncflags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	ch->tty = 0;
	wake_up_interruptible(&ch->open_wait);
}


static int pcxe_write(struct tty_struct * tty, int from_user,
		       unsigned char *buf, int count)
{
	struct channel *ch;
	struct board_chan *bc;
	int	c=0;
	int total=0, remain;
	unsigned int head, tail;
	unsigned long flags;

	/* printk("Entering pcxe_write()\n"); */

	ch = (struct channel *) tty->driver_data;

	bc = ch->brdchan;


	save_flags(flags);
	cli();

	while(1) {
		c = MIN(count, 1024);

		globalwinon(ch);

		head = bc->tin & (ch->txbufsize - 1);
		tail = bc->tout;
		tail &= (ch->txbufsize - 1);

		remain = tail - head - 1;
		if(remain < 0 )
			remain += ch->txbufsize;

		if(remain <= 0 || c <= 0)
			break;
		
		if(c > remain)
			c = remain;

		if(from_user) {
			memcpy_fromfs(pcxx_buf, buf, c);
			globalwinon(ch);
			data_to_board(ch, (unchar *) pcxx_buf, c);
			memoff(ch);
		} else {
			globalwinon(ch);
			data_to_board(ch, (unchar *) buf, c);
			memoff(ch);
		}

		buf += c;
		total += c;
		count -= c;
	}

	globalwinon(ch);

	/* 
		opost() sometimes needs more than 1 character...so I'll
		allow for about 10 or so.  That way, if write_chan()
		goes to sleep because opost() returned -1--which means
		there was no write room--I'll know enough to wake it up
		later.
	*/

	if(remain <= 10 && !(ch->statusflags & LOWWAIT)) {
		ch->statusflags |= LOWWAIT;
		bc->ilow = 1;
	}

	memoff(ch);
	restore_flags(flags);
	return(total);
}


static void pcxe_put_char(struct tty_struct *tty, unsigned char c)
{
	pcxe_write(tty, 0, &c, 1);
	return;
}


static int pcxe_write_room(struct tty_struct *tty)
{
	struct channel *ch;
	struct board_chan *bc;
	int remain;
	unsigned int head, tail;
	unsigned long flags;

	if(tty->stopped) {
		return(0);
	}

	ch = (struct channel *) tty->driver_data;
	bc = ch->brdchan;

	save_flags(flags);
	cli();

	globalwinon(ch);

	head = bc->tin & (ch->txbufsize - 1);
	tail = bc->tout;
	tail &= (ch->txbufsize - 1);

	if((remain = tail - head - 1) < 0 )
			remain += ch->txbufsize;

	memoff(ch);
	restore_flags(flags);

	return(remain);
}


static int pcxe_chars_in_buffer(struct tty_struct *tty)
{
	int chars;
	unsigned int ctail, head, tail;
	int remain;
	unsigned long flags;
	struct channel *ch = (struct channel *) tty->driver_data;
	struct board_chan *bc;

	/*
	 * Sigh...I have to check if driver_data is NULL here, because
	 * if an open() fails, the TTY subsystem eventually calls
	 * wait_until_sent(), which calls the driver's chars_in_buffer()
	 * routine.  And since the open() failed, we return 0 here.  TDJ
	 */

	if(ch == NULL)
		return(0);

	save_flags(flags);
	cli();
	globalwinon(ch);

	bc = ch->brdchan;

	tail = bc->tout;
	head = bc->tin;
	ctail = ch->mailbox->cout;
	if(tail == head && ch->mailbox->cin == ctail && bc->tbusy == 0)
		chars = 0;
	else {
		head = bc->tin & (ch->txbufsize - 1);
		tail &= (ch->txbufsize - 1);
		if((remain = tail - head - 1) < 0 )
			remain += ch->txbufsize;

		chars = (int)(ch->txbufsize - remain);

		/* 
		 * Make it possible to wakeup anything waiting for output
		 * in tty_ioctl.c, etc.
		 */
		if(!(ch->statusflags & EMPTYWAIT))
			setup_empty_event(tty);
	}

	memoff(ch);
	restore_flags(flags);

	return(chars);
}


static void pcxe_flush_buffer(struct tty_struct *tty)
{
	unsigned int tail;
	struct board_chan *bc;
	struct channel *ch;
	unsigned long flags;

	ch = (struct channel*) tty->driver_data;

	if(ch == NULL || (bc=ch->brdchan) == NULL) {
		console_print("ch or bc is NULL in pcxe_flush_buffer\n");
		return;
	}

	save_flags(flags);
	cli();

	globalwinon(ch);

	tail = bc->tout;
	fepcmd(ch, STOUT, (unsigned) tail, 0, 0, 0);

	memoff(ch);
	restore_flags(flags);

	wake_up_interruptible(&tty->write_wait);
	if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

static void pcxe_flush_chars(struct tty_struct *tty)
{
	/* 
	 * Don't think I need this, because this is called to empty the TX
	 * buffer for the 16450, 16550, etc.
	 */
}

static int block_til_ready(struct tty_struct *tty,struct file *filp,struct channel *info)
{
	struct wait_queue wait = {current, NULL};
	int	retval;
	int	do_clocal = C_CLOCAL(tty);

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if(info->asyncflags & ASYNC_CLOSING) {
		console_print("flags have ASYNC_CLOSING, sleeping on close_wait\n");
		interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
		if(info->asyncflags & ASYNC_HUP_NOTIFY)
			return -EAGAIN;
		else
			return -ERESTARTSYS;
#else
		return -EAGAIN;
#endif
	}

	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */
	if(tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if(info->asyncflags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;
		if((info->asyncflags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->asyncflags & ASYNC_SESSION_LOCKOUT) &&
		    (info->session != current->session))
		    return -EBUSY;
		if((info->asyncflags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->asyncflags & ASYNC_PGRP_LOCKOUT) &&
		    (info->pgrp != current->pgrp))
		    return -EBUSY;
		if(info->asyncflags & ASYNC_CALLOUT_ACTIVE)
			return(0);
	}
	
	/*
	 * If non-blocking mode is set, then make the check up front
	 * and then exit.
	 */
	if(filp->f_flags & O_NONBLOCK && tty->driver.subtype != SERIAL_TYPE_CALLOUT) {
		if(info->asyncflags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;
		info->asyncflags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}


	/*
	 * Block waiting for the carrier detect and the line to become free
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready before block: ttys%d, count = %d\n",
	       info->line, info->count);
#endif
	info->count--;
	info->blocked_open++;
	while(1) {
		current->state = TASK_INTERRUPTIBLE;
		if(tty_hung_up_p(filp) ||
		    !(info->asyncflags & ASYNC_INITIALIZED)) {
#ifdef SERIAL_DO_RESTART
			if(info->asyncflags & ASYNC_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;	
#else
			retval = -EAGAIN;
#endif
			break;
		}

		if(!(info->asyncflags & ASYNC_CALLOUT_ACTIVE) &&
				    !(info->asyncflags & ASYNC_CLOSING) && (do_clocal ||
											(info->imodem & info->dcd)))
			break;

		if(current->signal & ~current->blocked) {
			retval = -ERESTARTSYS;
			break;
		}

		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);
	if(!tty_hung_up_p(filp))
		info->count++;
	info->blocked_open--;
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready after blocking: ttys%d, count = %d\n",
	       info->line, info->count);
#endif
	if(retval)
		return retval;

	if(tty->driver.subtype == SERIAL_TYPE_CALLOUT)
		info->asyncflags |= ASYNC_CALLOUT_ACTIVE;
	else
		info->asyncflags |= ASYNC_NORMAL_ACTIVE;

	return 0;
}	


int pcxe_open(struct tty_struct *tty, struct file * filp)
{
	unsigned long flags;
	struct channel *ch;
	struct board_chan *bc;
	int line;
	int boardnum;
	int retval;
	unsigned int head;

	line = MINOR(tty->device) - tty->driver.minor_start;
	if(line < 0 || line >= NBDEVS) {
		console_print("line out of range in pcxe_open\n");
		tty->driver_data = NULL;
		return(-ENODEV);
	}

	boardnum = line / 16;
	if(boardnum >= NUMCARDS || boards[boardnum].status == DISABLED ||
								(line % MAXPORTS) >= boards[boardnum].numports) {
		tty->driver_data = NULL;   /* Mark this device as 'down' */
		return(-ENODEV);
	}
	
	ch = &digi_channels[line];

	down(&pcxx_sem);

	if(!pcxx_buf)
	{
		pcxx_buf = (unsigned char *) get_free_page(GFP_KERNEL);
		if(!pcxx_buf) {
			up(&pcxx_sem);
			return -ENOMEM;
		}
	}

	up(&pcxx_sem);
	
	if((bc=ch->brdchan) == 0) {
		tty->driver_data = NULL;
		return(-ENODEV);
	}

	ch->count++;
	tty->driver_data = ch;
	
	if(ch->count == 1 && (ch->asyncflags & ASYNC_SPLIT_TERMIOS)) {
		if(tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = ch->normal_termios;
		else 
			*tty->termios = ch->callout_termios;
	}

	ch->session = current->session;
	ch->pgrp = current->pgrp;

	save_flags(flags);
	cli();
	globalwinon(ch);
	ch->statusflags = 0;

	ch->imodem = bc->mstat;
	head = bc->rin;
	bc->rout = head;

	ch->tty = tty;

	pcxxparam(tty);

	ch->asyncflags |= ASYNC_INITIALIZED;

	memoff(ch);
	restore_flags(flags);

	retval = block_til_ready(tty, filp, ch);
	if(retval)
		return retval;

	/*
	 * Set this again in case a hangup set it to zero while this open()
	 * was waiting for the line...
	 */
	ch->tty = tty;

	save_flags(flags);
	cli();
	globalwinon(ch);
	bc->idata = 1;
	memoff(ch);
	restore_flags(flags);

	return 0;
} 


long pcxe_init(long kmem_start)
{
	ulong flags, save_loops_per_sec, memory_seg=0, memory_size;
	int lowwater, i, crd, shrinkmem=0, topwin = 0xff00L, botwin=0x100L;
	unchar *fepos, *memaddr, *bios, v;
	struct global_data *gd;
	struct board_info *bd;
	struct board_chan *bc;
	struct channel *ch;
	
	bh_base[DIGI_BH].routine = do_pcxe_bh;
	memset(&pcxe_driver, 0, sizeof(struct tty_driver));
	memset(&pcxe_callout, 0, sizeof(struct tty_driver));
	pcxe_driver.magic = TTY_DRIVER_MAGIC;
	pcxe_driver.name = "ttyd";
	pcxe_driver.major = DIGIMAJOR; 
	pcxe_driver.minor_start = 0;
	pcxe_driver.num = NBDEVS;
	pcxe_driver.type = TTY_DRIVER_TYPE_SERIAL;
	pcxe_driver.subtype = SERIAL_TYPE_NORMAL;
	pcxe_driver.init_termios = tty_std_termios;
	pcxe_driver.init_termios.c_iflag = 0;
	pcxe_driver.init_termios.c_oflag = 0;
	pcxe_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | CLOCAL | HUPCL;
	pcxe_driver.init_termios.c_lflag = 0;
	pcxe_driver.flags = TTY_DRIVER_REAL_RAW;
	pcxe_driver.refcount = &pcxe_refcount;
	pcxe_driver.table = pcxe_table;
	pcxe_driver.termios = pcxe_termios;
	pcxe_driver.termios_locked = pcxe_termios_locked;

	pcxe_driver.open = pcxe_open;
	pcxe_driver.close = pcxe_close;
	pcxe_driver.write = pcxe_write;
	pcxe_driver.write_room = pcxe_write_room;
	pcxe_driver.flush_buffer = pcxe_flush_buffer;
	pcxe_driver.chars_in_buffer = pcxe_chars_in_buffer;
	pcxe_driver.flush_chars = pcxe_flush_chars;
	pcxe_driver.put_char = pcxe_put_char;
	pcxe_driver.ioctl = pcxe_ioctl;
	pcxe_driver.set_termios = pcxe_set_termios;
	pcxe_driver.stop = pcxe_stop;
	pcxe_driver.start = pcxe_start;
	pcxe_driver.hangup = pcxe_hangup;
	pcxe_callout = pcxe_driver;
	pcxe_callout.name = "ttyD";
	pcxe_callout.major = DIGICUMAJOR;
	pcxe_callout.minor_start = 0;
	pcxe_callout.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL;
	pcxe_callout.subtype = SERIAL_TYPE_CALLOUT;

	pcxx_buf = 0;

	/* 
	 * loops_per_sec hasn't been set at this point :-(, so fake it out... 
	 * I set it so that I can use the __delay() function.
	 */
	save_loops_per_sec = loops_per_sec;
	loops_per_sec = 13L*500000L;

	save_flags(flags);
	cli();

	for(crd=0; crd < NUMCARDS; crd++) {
		bd = &boards[crd];
		outb(FEPRST, bd->port);
		pcxxdelay(1);

		for(i=0; (inb(bd->port) & FEPMASK) != FEPRST; i++) {
			if(i > 10000) {
				printk("Board not found at port 0x%x! Check switch settings.\n",
					bd->port);
				bd->status = DISABLED;
				break;
			}
			pcxxdelay(1);
		}
		if(bd->status == DISABLED)
			continue;

		v = inb(bd->port);

		if((v & 0x1) == 0x1) {
			if((v & 0x30) == 0) {        /* PC/Xi 64K card */
				memory_seg = 0xf000;
				memory_size = 0x10000;
			} 

			if((v & 0x30) == 0x10) {     /* PC/Xi 128K card */
				memory_seg = 0xe000;
				memory_size = 0x20000;
			}
			
			if((v & 0x30) == 0x20) {     /* PC/Xi 256K card */
				memory_seg = 0xc000;
				memory_size = 0x40000;
			}

			if((v & 0x30) == 0x30) {     /* PC/Xi 512K card */
				memory_seg = 0x8000;
				memory_size = 0x80000;
			}
			bd->type = PCXI;
		} else {
			if((v & 0x1) == 0x1) {
				bd->status = DISABLED;   /* PC/Xm unsupported card */
				printk("PC/Xm at 0x%x not supported!!\n", bd->port);
				continue;
			} else {
				if(v & 0xC0) {    
					topwin = 0x1f00L;
					outb((((ulong)bd->membase>>8) & 0xe0) | 0x10, bd->port+2);
					outb(((ulong)bd->membase>>16) & 0xff, bd->port+3);
					bd->type = PCXEVE; /* PC/Xe 8K card */
				} else { 
					bd->type = PCXE;    /* PC/Xe 64K card */
				}
					
				memory_seg = 0xf000;
				memory_size = 0x10000;
			}
		}

		memaddr = (unchar *) bd->membase;

		outb(FEPRST|FEPMEM, bd->port);

		for(i=0; (inb(bd->port) & FEPMASK) != (FEPRST|FEPMEM); i++) {
			if(i > 1000) {
				printk("%s not resetting at port 0x%x! Check switch settings.\n",
					board_desc[bd->type], bd->port);
				bd->status = DISABLED;
				break;
			}
			pcxxdelay(1);
		}
		if(bd->status == DISABLED)
			continue;

		memwinon(bd,0);
		*(ulong *)(memaddr + botwin) = 0xa55a3cc3;
		*(ulong *)(memaddr + topwin) = 0x5aa5c33c;

		if(*(ulong *)(memaddr + botwin) != 0xa55a3cc3 ||
					*(ulong *)(memaddr + topwin) != 0x5aa5c33c) {
			printk("Failed memory test at %lx for %s at port %x, check switch settings.\n",
				bd->membase, board_desc[bd->type], bd->port);
			bd->status = DISABLED;
			continue;
		}

		for(i=0; i < 16; i++) {
			memaddr[MISCGLOBAL+i] = 0;
		}

		if(bd->type == PCXI || bd->type == PCXE) {
			bios = memaddr + BIOSCODE + ((0xf000 - memory_seg) << 4);

			memcpy(bios, pcxx_bios, pcxx_nbios);

			outb(FEPMEM, bd->port);

			for(i=0; i <= 10000; i++) {
				if(*(ushort *)((ulong)memaddr + MISCGLOBAL) == *(ushort *)"GD" ) {
					goto load_fep;
				}
				pcxxdelay(1);
			}

			printk("BIOS download failed on the %s at 0x%x!\n",
							board_desc[bd->type], bd->port);
			bd->status = DISABLED;
			continue;
		}

		if(bd->type == PCXEVE) {
			bios = memaddr + (BIOSCODE & 0x1fff);
			memwinon(bd,0xff);
			
			memcpy(bios, pcxx_bios, pcxx_nbios);

			outb(FEPCLR, bd->port);
			memwinon(bd,0);

			for(i=0; i <= 10000; i++) {
				if(*(ushort *)((ulong)memaddr + MISCGLOBAL) == *(ushort *)"GD" ) {
					goto load_fep;
				}
				pcxxdelay(1);
			}

			printk("BIOS download failed on the %s at 0x%x!\n",
				board_desc[bd->type], bd->port);
			bd->status = DISABLED;
			continue;
		}

load_fep:
		fepos = memaddr + FEPCODE;
		if(bd->type == PCXEVE)
			fepos = memaddr + (FEPCODE & 0x1fff);

		memwinon(bd, (FEPCODE >> 13));
		memcpy(fepos, pcxx_cook, pcxx_ncook);
		memwinon(bd, 0);

		*(ushort *)((ulong)memaddr + MBOX +  0) = 2;
		*(ushort *)((ulong)memaddr + MBOX +  2) = memory_seg + FEPCODESEG;
		*(ushort *)((ulong)memaddr + MBOX +  4) = 0;
		*(ushort *)((ulong)memaddr + MBOX +  6) = FEPCODESEG;
		*(ushort *)((ulong)memaddr + MBOX +  8) = 0;
		*(ushort *)((ulong)memaddr + MBOX + 10) = pcxx_ncook;

		outb(FEPMEM|FEPINT, bd->port);
		outb(FEPMEM, bd->port);

		for(i=0; *(ushort *)((ulong)memaddr + MBOX); i++) {
			if(i > 2000) {
				printk("Command failed for the %s at 0x%x!\n",
					board_desc[bd->type], bd->port);
				bd->status = DISABLED;
				break;
			}
			pcxxdelay(1);
		}

		if(bd->status == DISABLED)
			continue;

		*(ushort *)(memaddr + FEPSTAT) = 0;
		*(ushort *)(memaddr + MBOX + 0) = 1;
		*(ushort *)(memaddr + MBOX + 2) = FEPCODESEG;
		*(ushort *)(memaddr + MBOX + 4) = 0x4L;

		outb(FEPINT, bd->port);
		outb(FEPCLR, bd->port);
		memwinon(bd, 0);

		for(i=0; *(ushort *)((ulong)memaddr + FEPSTAT) != *(ushort *)"OS"; i++) {
			if(i > 10000) {
				printk("FEP/OS download failed on the %s at 0x%x!\n",
					board_desc[bd->type], bd->port);
				bd->status = DISABLED;
				break;
			}
			pcxxdelay(1);
		}
		if(bd->status == DISABLED)
			continue;

		ch = &digi_channels[MAXPORTS * crd];
		pcxxassert(ch <= &digi_channels[NBDEVS-1], "ch out of range");

		bc = (struct board_chan *)((ulong)memaddr + CHANSTRUCT);
		gd = (struct global_data *)((ulong)memaddr + GLOBAL);

		if((bd->type == PCXEVE) && (*(ushort *)((ulong)memaddr+NPORT) < 3))
			shrinkmem = 1;

		for(i=0; i < bd->numports; i++, ch++, bc++) {
			if(((ushort *)((ulong)memaddr + PORTBASE))[i] == 0) {
				ch->brdchan = 0;
				continue;
			}
			ch->brdchan = bc;
			ch->mailbox = gd;
			ch->tqueue.routine = do_softint;
			ch->tqueue.data = ch;
			ch->board = &boards[crd];

			if(boards[crd].altpin) {
				ch->dsr = CD;
				ch->dcd = DSR;
				ch->digiext.digi_flags |= DIGI_ALTPIN;
			} else { 
				ch->dcd = CD;
				ch->dsr = DSR;
			}

			ch->boardnum = crd;

			ch->dev = (MAXPORTS * crd) + i;
			ch->tty = 0;

			if(shrinkmem) {
				fepcmd(ch, SETBUFFER, 32, 0, 0, 0);
				shrinkmem = 0;
			}
			
			if(bd->type != PCXEVE) {
				ch->txptr = memaddr+((bc->tseg-memory_seg) << 4);
				ch->rxptr = memaddr+((bc->rseg-memory_seg) << 4);
				ch->txwin = ch->rxwin = 0;
			} else {
				ch->txptr = memaddr+(((bc->tseg-memory_seg) << 4) & 0x1fff);
				ch->txwin = FEPWIN | ((bc->tseg-memory_seg) >> 9);
				ch->rxptr = memaddr+(((bc->rseg-memory_seg) << 4) & 0x1fff);
				ch->rxwin = FEPWIN | ((bc->rseg-memory_seg) >>9 );
			}

			ch->txbufhead = 0;
			ch->txbufsize = bc->tmax + 1;

			ch->rxbufhead = 0;
			ch->rxbufsize = bc->rmax + 1;

			lowwater = ch->txbufsize >= 2000 ? 1024 : ch->txbufsize/2;
			fepcmd(ch, STXLWATER, lowwater, 0, 10, 0);
			fepcmd(ch, SRXLWATER, ch->rxbufsize/4, 0, 10, 0);
			fepcmd(ch, SRXHWATER, 3 * ch->rxbufsize/4, 0, 10, 0);

			bc->edelay = 100;
			bc->idata = 1;

			ch->startc = bc->startc;
			ch->stopc = bc->stopc;
			ch->startca = bc->startca;
			ch->stopca = bc->stopca;

			ch->fepcflag = 0;
			ch->fepiflag = 0;
			ch->fepoflag = 0;
			ch->fepstartc = 0;
			ch->fepstopc = 0;
			ch->fepstartca = 0;
			ch->fepstopca = 0;

			ch->close_delay = 50;
			ch->count = 0;
			ch->blocked_open = 0;
			ch->callout_termios = pcxe_callout.init_termios;
			ch->normal_termios = pcxe_driver.init_termios;
			ch->open_wait = 0;
			ch->close_wait = 0;
		}

		sprintf(mesg, 
			"DigiBoard PC/Xx Driver V%s:  %s I/O=0x%x Mem=0x%lx Ports=%d\n", 
				VERSION, board_desc[bd->type], bd->port, bd->membase, bd->numports);
		console_print(mesg);

		memwinoff(bd, 0);
	}

	if(tty_register_driver(&pcxe_driver))
		panic("Couldn't register PC/Xe driver");

	if(tty_register_driver(&pcxe_callout))
		panic("Couldn't register PC/Xe callout");


	loops_per_sec = save_loops_per_sec;  /* reset it to what it should be */

	/*
	 * Start up the poller to check for events on all enabled boards
	 */
	pcxe_timer.function = pcxxpoll;
	pcxe_timer.expires  = HZ/25;
	add_timer(&pcxe_timer);

	restore_flags(flags);

	return(kmem_start);
}


static void pcxxpoll(unsigned long ignored)
{
	unsigned long flags;
	int crd;
	volatile unsigned int head, tail;
	struct channel *ch;
	struct board_info *bd;

	save_flags(flags);
	cli();
	del_timer(&pcxe_timer); /* Not that I think it matters with the ints off */

	for(crd=0; crd < NUMCARDS; crd++) {
		bd = &boards[crd];
		ch = &digi_channels[MAXPORTS*crd];

		if(bd->status == DISABLED)
			continue;

		assertmemoff(ch);

		globalwinon(ch);
		head = ch->mailbox->ein;
		tail = ch->mailbox->eout;

		if(head != tail)
			doevent(crd);

		memoff(ch);
	}

	pcxe_timer.function = pcxxpoll;
	pcxe_timer.expires  = HZ/25;
	add_timer(&pcxe_timer);
	restore_flags(flags);
}


static void doevent(int crd)
{
	struct board_info *bd;
	struct channel *ch;
	static struct tty_struct *tp;
	struct board_chan *bc;
	struct channel *chan0;
	unchar *eventbuf;
	volatile unsigned int head;
	volatile unsigned int tail;
	int channel, event, mstat, lstat;

	bd = &boards[crd];

	chan0 = &digi_channels[MAXPORTS * crd];
	pcxxassert(chan0 <= &digi_channels[NBDEVS-1], "ch out of range");

	assertgwinon(chan0);

	head = chan0->mailbox->ein;
	tail = chan0->mailbox->eout;


	if(head >= IMAX-ISTART || tail >= IMAX-ISTART || (head|tail) & 03) {
		console_print("head or tail out of range in doevent!\n");
		return;
	}

	while(tail != head) {
		assertgwinon(chan0);
		eventbuf = bd->membase + tail + ISTART;
		channel = eventbuf[0];
		event = eventbuf[1];
		mstat = eventbuf[2];
		lstat = eventbuf[3];

		ch=chan0+channel;
		bc = ch->brdchan;
		if(!(event & (MODEMCHG_IND|BREAK_IND|LOWTX_IND|EMPTYTX_IND|DATA_IND))
						|| (unsigned)channel >= bd->numports || !bc || !ch) { 
			sprintf(mesg, "physmem=%lx, tail=%x, head=%x\n", bd->membase, 
																tail, head);
			console_print(mesg);
			sprintf(mesg,
					"doevent(%x) channel %x, event %x, mstat %x, lstat %x, bc=%x\n",
					crd, (unsigned)channel, event, (unsigned)mstat, lstat, bc);
			console_print(mesg);

			if(channel >= bd->numports) {
				ch = chan0;
				bc = ch->brdchan;
			}
			goto next;
		}

		if(event & DATA_IND) {
			receive_data(ch);
			assertgwinon(ch);
		}

		if(event & MODEMCHG_IND) {
			ch->imodem = mstat;
/*
			sprintf(mesg, "Modem signal change, mstat=%x\n", mstat);
			console_print(mesg);
*/
			if(ch->asyncflags & ASYNC_CHECK_CD) {
				if(mstat & ch->dcd)
					wake_up_interruptible(&ch->open_wait);
				else
					pcxe_sched_event(ch, PCXE_EVENT_HANGUP);
			}
		}

		if(event & BREAK_IND) {
			tp = ch->tty;
			if(!tp)
				break;

			tp->flip.count++;
			*tp->flip.flag_buf_ptr++ = TTY_BREAK;

			/* 
			 * Right now, there is no do_SAK() handling here, since
			 * there is no 'set ASYNC_SAK' ioctl() yet...I'll do it
			 * later.
			 */

			/* Must be called with global data */
			tty_schedule_flip(ch->tty); 
		}

		if(event & LOWTX_IND) {
			if(ch->statusflags & LOWWAIT) {
				tp = ch->tty;
				if(!tp)
					break;
				if(!tp->stopped) {
					ch->statusflags &= ~LOWWAIT;
					wake_up_interruptible(&tp->write_wait);
				}
			}
		}

		if(event & EMPTYTX_IND) {
			tp = ch->tty;
			if(!tp)
				break;

			if(ch->statusflags & EMPTYWAIT) {
				ch->statusflags &= ~EMPTYWAIT;
				wake_up_interruptible(&tp->write_wait);
			}
		}

	next:
		if(!bc)
			console_print("bc == NULL in doevent!\n");
		else
			bc->idata = 1;
		tail = (tail+4) & (IMAX-ISTART-4);

		globalwinon(chan0);
	}

	chan0->mailbox->eout = tail;
	return;
}


/*
 * pcxxdelay - delays a specified number of milliseconds
 */
static void pcxxdelay(int msec)
{
	while(msec-- > 0)
		__delay(loops_per_sec/1000);
}


static void 
fepcmd(struct channel *ch, int cmd, int word_or_byte, int byte2, int ncmds,
						int bytecmd)
{
	unchar *memaddr;
	unsigned int head, tail;
	long count;
	int n;

	if(ch->board->status == DISABLED)
		return;

	memaddr = ch->board->membase;
	head = ch->mailbox->cin;

	if(head >= (CMAX-CSTART) || (head & 03)) {
		sprintf(mesg, "line %d: Out of range, cmd=%x, head=%x\n", __LINE__, 
						cmd, head);
		console_print(mesg);
		return;
	}

	if(bytecmd) {
		*(unchar *)(memaddr+head+CSTART+0) = cmd;
		*(unchar *)(memaddr+head+CSTART+1) = PORTNUM(ch);
		*(unchar *)(memaddr+head+CSTART+2) = word_or_byte;
		*(unchar *)(memaddr+head+CSTART+3) = byte2;
	} else {
		*(unchar *)(memaddr+head+CSTART+0) = cmd;
		*(unchar *)(memaddr+head+CSTART+1) = PORTNUM(ch);
		*(ushort*)(memaddr+head+CSTART+2) = word_or_byte;
	}

	head = (head+4) & (CMAX-CSTART-4);
	ch->mailbox->cin = head;

	count = FEPTIMEOUT;

	while(1) {
		count--;
		if(count == 0) {
			console_print("Fep not responding in fepcmd()\n");
			return;
		}

		head = ch->mailbox->cin;
		tail = ch->mailbox->cout;

		n = (head-tail) & (CMAX-CSTART-4);

		if(n <= ncmds * (sizeof(short)*4))
			break;
	}
}


static void data_to_board(struct channel *ch, unchar *buf, int cnt)
{
	int n;
	unchar *txptr;
	struct board_chan *bc;
	unsigned int head;

	bc = ch->brdchan;
	pcxxassert(bc != 0, "bc out of range");

	head = bc->tin;

	if(cnt > ch->txbufsize || (unsigned)(head-ch->txbufhead) >= ch->txbufsize) {
		sprintf(mesg, "head = %x, txbufhead=%x, cnt=%x\n", head, ch->txbufhead, cnt);
		console_print(mesg);
		pcxxassert(0, "pointer out of range");
		return;
	}

	n = ch->txbufhead + ch->txbufsize - head;

	pcxxassert(n > 0, "n out of range");

	txwinon(ch);

	if(cnt >= n) {
		cnt -= n;
		txptr = ch->txptr + head;
		while(--n >= 0)
			*txptr++ = *buf++;
		head = ch->txbufhead;
	}

	txptr = ch->txptr + head;
	n = cnt;
	while(--n >= 0)
		*txptr++ = *buf++;

	head += cnt;

	globalwinon(ch);
	bc->tin = head;
	memoff(ch);
}


static void pcxxparam(struct tty_struct *tty)
{
	struct channel *ch;
	struct board_chan *bc;
	unsigned int head;
	int mval, hflow, cflag, iflag;
	struct termios *ts;

	ch = (struct channel *)tty->driver_data;
	bc = ch->brdchan;
	pcxxassert(bc != 0, "bc out of range");
	pcxxassert(ch != 0, "ch out of range");

	ts = tty->termios;

	if((ts->c_cflag & CBAUD) == 0) {
		head = bc->rin;
		bc->rout = head;
		head = bc->tin;
		fepcmd(ch, STOUT, (unsigned) head, 0, 0, 0);
		mval = 0;
	} else {
		cflag = ts->c_cflag & (PARODD|PARENB|CSTOPB|CSIZE|CBAUD);

		/*
		 * HUPCL bit is used by FEP to indicate fast baud
		 * table is to be used.
		 */
		if(ch->digiext.digi_flags & DIGI_FAST)
			cflag |= HUPCL;

		if(cflag != ch->fepcflag) {
			ch->fepcflag = cflag;
			fepcmd(ch, SETCTRLFLAGS, (unsigned) cflag, 0, 0, 0);
		}

		if(cflag & CLOCAL || tty->driver.subtype != SERIAL_TYPE_CALLOUT)
			ch->asyncflags &= ~ASYNC_CHECK_CD;
		else
			ch->asyncflags |= ASYNC_CHECK_CD;

		mval = DTR | RTS;
	}


	iflag = ts->c_iflag &
		(IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK|ISTRIP|IXON|IXANY|IXOFF);
	
	if(ch->digiext.digi_flags & DIGI_AIXON)
		iflag |= IAIXON;

	if(iflag != ch->fepiflag) {
		ch->fepiflag = iflag;
		fepcmd(ch, SETIFLAGS, (unsigned int) ch->fepiflag, 0, 0, 0);
	}

	if((ts->c_cflag & CLOCAL) || (ch->digiext.digi_flags & DIGI_FORCEDCD)) {
		if(ch->digiext.digi_flags & DIGI_FORCEDCD)
			bc->mint = 0;
		else
			bc->mint = ch->dcd;
	} else
		bc->mint = ch->dcd;

	ch->imodem = bc->mstat;

	hflow = 0;

	if(ts->c_cflag & CRTSCTS) 
		hflow |= (CTS | RTS);

	if(ch->digiext.digi_flags & RTSPACE)
		hflow |= RTS;
	if(ch->digiext.digi_flags & DTRPACE)
		hflow |= DTR;
	if(ch->digiext.digi_flags & CTSPACE)
		hflow |= CTS;
	if(ch->digiext.digi_flags & DSRPACE)
		hflow |= ch->dsr;
	if(ch->digiext.digi_flags & DCDPACE)
		hflow |= ch->dcd;

	if(hflow != ch->hflow) {
		ch->hflow = hflow;
		fepcmd(ch, SETHFLOW, hflow, 0xff, 0, 1);
	}

	mval ^= ch->modemfake & (mval ^ ch->modem);

	if(ch->omodem ^ mval) {
		ch->omodem = mval;
		fepcmd(ch, SETMODEM, mval, RTS|DTR, 0, 1);
	}

	if(ch->startc != ch->fepstartc || ch->stopc != ch->fepstopc) {
		ch->fepstartc = ch->startc;
		ch->fepstopc = ch->stopc;
		fepcmd(ch, SONOFFC, ch->fepstartc, ch->fepstopc, 0, 1);
	}

	if(ch->startca != ch->fepstartca || ch->stopca != ch->fepstopca) {
		ch->fepstartca = ch->startca;
		ch->fepstopca = ch->stopca;
		fepcmd(ch, SAUXONOFFC, ch->fepstartca, ch->fepstopca, 0, 1);
	}
}


static void receive_data(struct channel *ch)
{
	struct board_chan *bc;
	struct tty_struct *tp;
	unsigned int tail, head, wrapmask;
	int n;
	int piece;
	struct termios *ts=0;
	unchar *rptr;
	int rc;
	int wrapgap;

	tp = ch->tty;

	if(tp)
		ts = tp->termios;

	bc = ch->brdchan;

	if(!bc) {
		console_print("bc is NULL in receive_data!\n");
		return;
	}

	wrapmask = ch->rxbufsize - 1;

	head = bc->rin;
	head &= wrapmask;
	tail = bc->rout & wrapmask;

	n = (head-tail) & wrapmask;

	if(n == 0)
		return;

	/*
	 * If CREAD bit is off or device not open, set TX tail to head
	 */
	if(!tp || !ts || !(ts->c_cflag & CREAD)) {
/*
		if(!tp || !ts)
			console_print("device is not open (tp=NULL)\n");
		else if(!(ts->c_cflag & CREAD))
			console_print("CREAD is off\n");
*/

		bc->rout = head;
		return;
	}

	if(tp->flip.count == TTY_FLIPBUF_SIZE) {
		console_print("tp->flip.count = TTY_FLIPBUF_SIZE\n");
		return;
	}

	if(bc->orun) {
		bc->orun = 0;
		sprintf(mesg, "overrun! DigiBoard device minor=%d\n",MINOR(tp->device));
		console_print(mesg);
		return;
	}

	while(1) {
		rxwinon(ch);
		rptr = tp->flip.char_buf_ptr;
		rc = tp->flip.count;
		while(n > 0) {
			if(head >= tail)
				wrapgap = head - tail;
			else
				wrapgap = ch->rxbufsize - tail;

			piece = (wrapgap < n) ? wrapgap : n;

			memcpy(rptr, ch->rxptr + tail, piece);

			rptr += piece;
			rc += piece;

			tail = (tail + piece) & wrapmask;
			n -= piece;
		}

		tp->flip.count = rc;

		tp->flip.char_buf_ptr = rptr;

	    globalwinon(ch);
		bc->rout = tail;
		break;
	}

	/* Must be called with global data */
	tty_schedule_flip(ch->tty); 

	return;
}


static int pcxe_ioctl(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned long arg)
{
	int error;
	struct channel *ch = (struct channel *) tty->driver_data;
	struct board_chan *bc;
	int retval;
	unsigned int mflag, mstat;
	unsigned char startc, stopc;
	unsigned long flags;
	digiflow_t dflow;

	if(ch)
		bc = ch->brdchan;
	else {
		console_print("ch is NULL in pcxe_ioctl!\n");
		return(-EINVAL);
	}

	save_flags(flags);

	switch(cmd) {
		case TCSBRK:	/* SVID version: non-zero arg --> no break */
			retval = tty_check_change(tty);
			if(retval)
				return retval;
			setup_empty_event(tty);		
			wait_until_sent(tty, 0);
			if(!arg)
				digi_send_break(ch, HZ/4);    /* 1/4 second */
			return 0;

		case TCSBRKP:	/* support for POSIX tcsendbreak() */
			retval = tty_check_change(tty);
			if(retval)
				return retval;
			setup_empty_event(tty);		
			wait_until_sent(tty, 0);
			digi_send_break(ch, arg ? arg*(HZ/10) : HZ/4);
			return 0;

		case TIOCGSOFTCAR:
			error = verify_area(VERIFY_WRITE, (void *) arg,sizeof(long));
			if(error)
				return error;
			put_fs_long(C_CLOCAL(tty) ? 1 : 0,
				    (unsigned long *) arg);
			return 0;

		case TIOCSSOFTCAR:
			arg = get_fs_long((unsigned long *) arg);
			tty->termios->c_cflag =
				((tty->termios->c_cflag & ~CLOCAL) |
				 (arg ? CLOCAL : 0));
			return 0;

		case TIOCMODG:
		case TIOCMGET:
			mflag = 0;

			cli();
			globalwinon(ch);
			mstat = bc->mstat;
			memoff(ch);
			restore_flags(flags);

			if(mstat & DTR)
				mflag |= TIOCM_DTR;
			if(mstat & RTS)
				mflag |= TIOCM_RTS;
			if(mstat & CTS)
				mflag |= TIOCM_CTS;
			if(mstat & ch->dsr)
				mflag |= TIOCM_DSR;
			if(mstat & RI)
				mflag |= TIOCM_RI;
			if(mstat & ch->dcd)
				mflag |= TIOCM_CD;

			error = verify_area(VERIFY_WRITE, (void *) arg,sizeof(long));
			if(error)
				return error;
			put_fs_long(mflag, (unsigned long *) arg);
			break;

		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMODS:
		case TIOCMSET:
			mstat = get_fs_long((unsigned long *) arg);

			mflag = 0;
			if(mstat & TIOCM_DTR)
				mflag |= DTR;
			if(mstat & TIOCM_RTS)
				mflag |= RTS;

			switch(cmd) {
				case TIOCMODS:
				case TIOCMSET:
					ch->modemfake = DTR|RTS;
					ch->modem = mflag;
					break;

				case TIOCMBIS:
					ch->modemfake |= mflag;
					ch->modem |= mflag;
					break;

				case TIOCMBIC:
					ch->modemfake |= mflag;
					ch->modem &= ~mflag;
					break;
			}

			cli();
			globalwinon(ch);
			pcxxparam(tty);
			memoff(ch);
			restore_flags(flags);
			break;

		case TIOCSDTR:
			ch->omodem |= DTR;
			cli();
			globalwinon(ch);
			fepcmd(ch, SETMODEM, DTR, 0, 10, 1);
			memoff(ch);
			restore_flags(flags);
			break;

		case TIOCCDTR:
			ch->omodem &= ~DTR;
			cli();
			globalwinon(ch);
			fepcmd(ch, SETMODEM, 0, DTR, 10, 1);
			memoff(ch);
			restore_flags(flags);
			break;

		case DIGI_GETA:
			if((error=
				verify_area(VERIFY_WRITE, (char*)arg, sizeof(digi_t))))
				return(error);

			memcpy_tofs((char*)arg, &ch->digiext, sizeof(digi_t));
			break;

		case DIGI_SETAW:
		case DIGI_SETAF:
			if(cmd == DIGI_SETAW) {
				setup_empty_event(tty);		
				wait_until_sent(tty, 0);
			}
			else {
				if(tty->ldisc.flush_buffer)
					tty->ldisc.flush_buffer(tty);
			}

			/* Fall Thru */

		case DIGI_SETA:
			if((error=
				verify_area(VERIFY_READ, (char*)arg,sizeof(digi_t))))
				return(error);

			memcpy_fromfs(&ch->digiext, (char*)arg, sizeof(digi_t));
			
			if(ch->digiext.digi_flags & DIGI_ALTPIN) {
				ch->dcd = DSR;
				ch->dsr = CD;
			} else {
				ch->dcd = CD;
				ch->dsr = DSR;
			}
		
			cli();
			globalwinon(ch);
			pcxxparam(tty);
			memoff(ch);
			restore_flags(flags);
			break;

		case DIGI_GETFLOW:
		case DIGI_GETAFLOW:
			cli();	
			globalwinon(ch);
			if(cmd == DIGI_GETFLOW) {
				dflow.startc = bc->startc;
				dflow.stopc = bc->stopc;
			} else {
				dflow.startc = bc->startca;
				dflow.stopc = bc->stopca;
			}
			memoff(ch);
			restore_flags(flags);

			if((error=verify_area(VERIFY_WRITE, (char*)arg,sizeof(dflow))))
				return(error);

			memcpy_tofs((char*)arg, &dflow, sizeof(dflow));
			break;

		case DIGI_SETAFLOW:
		case DIGI_SETFLOW:
			if(cmd == DIGI_SETFLOW) {
				startc = ch->startc;
				stopc = ch->stopc;
			} else {
				startc = ch->startca;
				stopc = ch->stopca;
			}

			if((error=verify_area(VERIFY_READ, (char*)arg,sizeof(dflow))))
				return(error);

			memcpy_fromfs(&dflow, (char*)arg, sizeof(dflow));

			if(dflow.startc != startc || dflow.stopc != stopc) {
				cli();
				globalwinon(ch);

				if(cmd == DIGI_SETFLOW) {
					ch->fepstartc = ch->startc = dflow.startc;
					ch->fepstopc = ch->stopc = dflow.stopc;
					fepcmd(ch,SONOFFC,ch->fepstartc,ch->fepstopc,0, 1);
				} else {
					ch->fepstartca = ch->startca = dflow.startc;
					ch->fepstopca  = ch->stopca = dflow.stopc;
					fepcmd(ch, SAUXONOFFC, ch->fepstartca, ch->fepstopca, 0, 1);
				}

				if(ch->statusflags & TXSTOPPED)
					pcxe_start(tty);

				memoff(ch);
				restore_flags(flags);
			}
			break;

		default:
			return -ENOIOCTLCMD;
	}

	return 0;
}

static void pcxe_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct channel *info = (struct channel *) tty->driver_data;
	unsigned long flags;

	save_flags(flags);
	cli();
	globalwinon(info);
	pcxxparam(tty);
	memoff(info);
	restore_flags(flags);

	if(!(old_termios->c_cflag & CLOCAL) && (tty->termios->c_cflag & CLOCAL))
		wake_up_interruptible(&info->open_wait);
}


static void do_pcxe_bh(void *unused)
{
	run_task_queue(&tq_pcxx);
}


static void do_softint(void *private_)
{
	struct channel *info = (struct channel *) private_;
	struct tty_struct *tty;
	
	tty = info->tty;
	if(!tty)
		return;

	if(clear_bit(PCXE_EVENT_HANGUP, &info->event)) {
		tty_hangup(tty);
		wake_up_interruptible(&info->open_wait);
		info->asyncflags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	}
}


static void pcxe_stop(struct tty_struct *tty)
{
	struct channel *info = (struct channel *)tty->driver_data;
	unsigned long flags;
	
	save_flags(flags); 
	cli();
	globalwinon(info);
	fepcmd(info, PAUSETX, 0, 0, 0, 0);
	info->statusflags |= TXSTOPPED;
	memoff(info);
	restore_flags(flags);
}


static void pcxe_start(struct tty_struct *tty)
{
	struct board_chan *bc;
	struct channel *info = (struct channel *)tty->driver_data;
	unsigned long flags;

	/* Just in case output was resumed because of a change in Digi-flow */
	if(!(info->statusflags & TXSTOPPED))
		return;
	
	save_flags(flags); cli();
	globalwinon(info);
	bc = info->brdchan;
	if(info->statusflags & LOWWAIT)
		bc->ilow = 1;

	fepcmd(info, RESUMETX, 0, 0, 0, 0);
	info->statusflags &= ~TXSTOPPED;
	memoff(info);
	restore_flags(flags);
}


void digi_send_break(struct channel *ch, int msec)
{
	unsigned long flags;

	save_flags(flags);
	cli();
	globalwinon(ch);

	/* 
	 * Maybe I should send an infinite break here, schedule() for
	 * msec amount of time, and then stop the break.  This way,
	 * the user can't screw up the FEP by causing digi_send_break()
	 * to be called (i.e. via an ioctl()) more than once in msec amount 
	 * of time.  Try this for now...
	 */

	fepcmd(ch, SENDBREAK, msec, 0, 10, 0);
	memoff(ch);

	restore_flags(flags);
}

void setup_empty_event(struct tty_struct *tty)
{
	struct channel *ch = tty->driver_data;
	struct board_chan *bc=ch->brdchan;
	unsigned long flags;

	save_flags(flags);
	cli();
	globalwinon(ch);
	ch->statusflags |= EMPTYWAIT;
	bc->iempty = 1;
	memoff(ch);
	restore_flags(flags);
}
